/****************************************************************************
**
** Copyright (C) Datafox GmbH.
** All rights reserved.
** Contact: Datafox GmbH (develop@datafox.de)
**
** This file is part of the DFComDLL module of the Datafox SDK.
**
****************************************************************************/
#pragma once
#ifndef DATAFOX_COMMUNICATION_MODULE_BASE_H
#define DATAFOX_COMMUNICATION_MODULE_BASE_H

// Protokollklasse einbinden.
#include "datafox/communication/protocol.h"

#include <boost/property_tree/ptree.hpp>
#include <boost/property_tree/json_parser.hpp>

#ifdef WITH_VIRTUAL_DEVICE
#include <boost/uuid/uuid.hpp>
#include <boost/uuid/uuid_generators.hpp>
#include <boost/uuid/uuid_io.hpp>

#include <boost/filesystem.hpp>
#include <string>
#endif

/** In diesem Module sind alle Basisspezifischen Modulfunktionen enthalten.
*
* 
*
*/

namespace Datafox
{
    namespace Communication
    {
        // Fuer die Verwendung der Verbindung im Active-Mode
        enum active_mode_state
        { 
            ams_unused = 0, // Verbindungsklasse ist nicht in Verwendung
            ams_starting,   // Verbindungsklasse wird vorbereitet.
            ams_accepting,  // Verbindungsklasse nimmt eine Verbindung entgegen.
            ams_detecting,  // Verbindungsklasse ermittelt Daten vom Geraet.
            ams_usable,     // Verbindungsklasse hat die zu ermittelten Daten vom Geraet erhalten.
            ams_binding,    // Verbindungsklasse durchläuft die Bindungsprüfung.
            ams_active      // Verbindungsklasse befindet sich im Active-Mode.
        };

        class Module_base
        {
        public:
            Module_base( int connection_id = Connection_base::dv_connection_id, uint8_t device_address = Protocol_base::dv_device_address );
            virtual ~Module_base();

        private:
            std::shared_ptr<Protocol> m_protocol;

        public:
            std::shared_ptr<Protocol> get_protocol_pointer()
            {
                return m_protocol;
            }
            void set_protocol_pointer( std::shared_ptr<Protocol> base )
            {
                m_protocol = base;
            }

            Protocol& get_protocol()
            {
                return *m_protocol.get();
            }
        private:
            std::shared_ptr<Trace_base> m_trace;

        public:
            std::shared_ptr<Trace_base> get_trace_pointer()
            {
                return m_trace;
            }
            void set_trace_pointer( std::shared_ptr<Trace_base> base )
            {
                m_trace = base;
            }

        public:
            Connection_base& get_connection()
            {
                return get_protocol().get_connection();
            }

            /// Oeffnen der Schnittstelle
            int open_connection( Connection_base::Connection_type type, const std::string& connection_string, uint32_t connection_value, uint32_t communication_read_timeout )
            {
                return get_connection().open( type, connection_string, connection_value, communication_read_timeout );
            }
            /// Schliessen der Schnittstelle.
            virtual int close_connection()
            {
                get_protocol().aes_crypt_session().reset_session_key();
                return get_connection().close();
            }

            bool is_connection_open()
            {
                return get_connection().is_open();
            }

        public:
            virtual void trace( const char *format... )
            {
                if ( m_trace.get() != nullptr && m_trace->get_mode() >= Trace_base::Trace_mode::on)
                {
                    va_list args;
                    va_start( args, format );
                    m_trace->print( format, args );
                    va_end( args );
                }
            }

            void set_trace_mode( Trace_base::Trace_mode mode = Trace_base::Trace_mode::on )
            {
                if ( m_trace.get() != nullptr )
                {
                    m_trace->set_mode( mode );
                }
            }
            Trace_base::Trace_mode get_trace_mode()
            {
                Trace_base::Trace_mode mode = Trace_base::Trace_mode::off;
                if ( m_trace.get() != nullptr )
                {
                    mode = m_trace->get_mode();
                }

                return mode;
            }

            void set_trace_file_path( const std::string &file_path )
            {
                if ( m_trace.get() != nullptr )
                {
                    m_trace->set_file_path( file_path );
                }
            }
            std::string get_trace_file_path()
            {
                std::string file_path;
                if ( m_trace.get() != nullptr )
                {
                    file_path = m_trace->get_file_path();
                }

                return file_path;
            }

        protected:
            bool m_auto_baud;

        public:
            // Autobauding mit prio 38400, 9600, 19200. ( HK hinzugefügt 57600, 115200 )
            bool get_auto_baud()
            {
                return m_auto_baud;
            }
            void set_auto_baud( bool val )
            {
                m_auto_baud = val;
            }

            void reset();

            // Fehlerbearbeitung
        public:
            void reset_error();

            // Fehlertexte fuer die Module/Funktionen
            uint16_t m_error_module_number;
            std::string m_error_module_text;
            virtual void error_module( uint16_t module, uint16_t error, std::string &text );

            uint16_t m_error_function_number;
            std::string m_error_function_text;
            virtual void error_function( uint16_t module, uint16_t func, uint16_t error, std::string &text );

            void trace_error( uint8_t device_address, uint16_t module, uint16_t func, int error );

            uint8_t m_packet_data[IV_PACKET_LENGTH_DATA]; // 21 Byte Paketframe

        private:
            volatile active_mode_state m_active_mode_state;

        public:
            // Verbindungsobjekt im Active-Mode Verwendet.
            inline void set_active_mode_state( active_mode_state state ) 
            { 
                m_active_mode_state = state; 
            }
            inline active_mode_state get_active_mode_state() 
            { 
                return m_active_mode_state; 
            }

        protected:
            std::recursive_mutex m_mutex_call; // Eine in sich geschlossene Kommunikation Packetaustausch. Inkl. Aktive-Antwort
            typedef std::lock_guard<std::recursive_mutex> lock_guard_call;

        protected:
            // Call-Back-Funktion
            std::function<int (unsigned int size, unsigned int count) > m_call_back_function_with_parameters;
        public:
            void set_call_back_function_with_parameters( std::function<int ( unsigned int, unsigned int )> func );

            //#################################
            // CallRoutinen
        public:
            // Ausfuehren eines Protokollbefehls
            virtual int call( uint16_t context, uint16_t module, uint16_t func, uint8_t *buffer, uint16_t &count, int context_value = -1 );
            // Ausfuehren von Befehlen
            virtual int remote_call( uint16_t module, uint16_t func, uint8_t *data, uint16_t &count );

        protected:
            int m_result;         // Returnwert der call-Routine

        public:
            inline int get_last_result()
            { 
                return m_result;
            }

            void set_connection_id( int id )
            {
                get_connection().set_id( id );
            }
            int get_connection_id()
            {
                return get_connection().get_id();
            }

            bool is_own_device_address( uint8_t device_address )
            {
                return get_protocol().is_own_device_address( device_address );
            }

            void set_device_address_receiver( uint8_t device_address )
            {
                get_protocol().set_device_address_receiver( device_address );
            }
            uint8_t get_device_address_receiver()
            {
                return get_protocol().get_device_address_receiver();
            }

            void set_device_address_sender( uint8_t device_address )
            {
                get_protocol().set_device_address_sender( device_address );
            }
            uint8_t get_device_address_sender()
            {
                return get_protocol().get_device_address_sender();
            }

            uint32_t get_crc32( uint8_t *data, std::size_t size, uint32_t crc = 0xFFFFFFFF, bool closure = true )
            {
                return get_protocol().get_crc32( data, size, crc, closure );
            }

        public:
            // Einfache Anfrage ob das Geraet antwortet.
            int ping( uint8_t device_address, uint8_t *data, uint16_t *count );

        public:
            template<class Type>
            Type get_property( const std::string &name, Type default_value )
            {
                return m_device_properties.get<Type>( name, default_value );
            }

            template<class Type>
            void set_property( const std::string &name, Type value )
            {
                m_device_properties.put<Type>( name, value );
            }

        private:
            boost::property_tree::ptree m_device_properties;

#ifdef WITH_VIRTUAL_DEVICE
        public:
            void set_uuid( boost::uuids::uuid &uuid );
            boost::uuids::uuid get_uuid();
            std::string get_uuid_string();
    
            void set_path_cache( const boost::filesystem::path &path );
            boost::filesystem::path get_path_cache();
    
            std::string get_path_cache_device();
            std::string get_file_path_device();

        protected:
            virtual void load_from_cache();
            virtual void save_to_cache();

        private:
            boost::filesystem::path m_path_cache;
            boost::uuids::uuid m_uuid;
#endif
        };

    } // namespace Communication
} // namespace Datafox

#endif
