/****************************************************************************
**
** Copyright (C) Datafox GmbH.
** All rights reserved.
** Contact: Datafox GmbH (develop@datafox.de)
**
** This file is part of the DFComDLL module of the Datafox SDK.
**
****************************************************************************/

#include "datafox/communication/module_base.h"
#include <boost/format.hpp>

// ##################################################################
/**
* @brief Initialisieren der Membervariablen.
*
* @param connection_id Nummer der Verbindungsklasse fuer z.B. Trace.
*/
Datafox::Communication::Module_base::Module_base( int connection_id, uint8_t device_address )
    : m_protocol( std::make_shared<Protocol>( connection_id, device_address ) )
    , m_trace( get_connection().get_trace_pointer() )
#ifdef WITH_VIRTUAL_DEVICE
    , m_uuid( boost::uuids::random_generator()() )
#endif
{
    // Zustand des Kanalobjekts
    m_active_mode_state = ams_unused;

    m_result = IV_ERROR_SUCCESS;

    // Callbacks intialisieren
    m_call_back_function_with_parameters = nullptr;
    get_protocol().remote_call = std::bind( &Module_base::remote_call, this, std::placeholders::_1, std::placeholders::_2, std::placeholders::_3, std::placeholders::_4 );

    // Kein Autobaud bei Verbindungsherstellung.
    set_auto_baud( false );
}

// ##################################################################
/**
* @brief Destruktor.
*
*/
Datafox::Communication::Module_base::~Module_base() 
{

}

// ##################################################################
/**
* @brief Initialisiert die grundlegend benoetigten Membervariablen.
*/
void Datafox::Communication::Module_base::reset()
{
    trace( "%s(%i):INFO(3): %s()\n", __FILE__, __LINE__, __func__ );

    m_result = IV_ERROR_SUCCESS;
    get_protocol().reset();
    std::fill_n( m_packet_data, sizeof( m_packet_data ), 0 );
}

// ##################################################################
/**
* @brief Setzen einer auszufuehrenden remoteCall-Back Funktion mit Parametern.
*/
void Datafox::Communication::Module_base::set_call_back_function_with_parameters( std::function<int ( unsigned int size, unsigned int count)> func )
{
    trace( "%s(%i):INFO(3): %s( %p )\n", __FILE__, __LINE__, __func__, func );

    m_call_back_function_with_parameters = func;
}


//#########################################################
//####################### ERROR ###########################
//#########################################################

// ##################################################################
/**
* @brief Die gesamte Fehlerbehandlung zuruecksetzen.
*/
void Datafox::Communication::Module_base::reset_error()
{
    trace( "%s(%i):INFO(3): %s()\n", __FILE__, __LINE__, __func__ );

    // Fehlernummern loeschen
    m_error_module_number = 0;
    m_error_function_number = 0;

    // Puffer die in Gebrauch waren, loeschen.
    m_error_module_text.clear();
    m_error_function_text.clear();

    // Reset des Fehlerzustandes der Protokollklasse
    get_protocol().reset_error();
}

// ##################################################################
/**
* @brief Beschreibung fehlt noch.
*/
void Datafox::Communication::Module_base::error_module( uint16_t module, uint16_t error, std::string &text )
{
    trace( "%s(%i):INFO(3): %s( %u, %u, %p )\n", __FILE__, __LINE__, __func__, module, error, text.c_str() );

    // Zuweisen eines modulübergreifenden Textes zum Modulfehler.
    switch ( error )
    {
    case 1:
        text = boost::str( boost::format( "( %u ): Das geforderte Modul wird nicht unterstützt." ) % error );
        break;

    case 2:
        text = boost::str( boost::format( "( %u ): Die geforderte Funktion wird nicht unterstützt." ) % error );
        break;

    case 3:
        text = boost::str( boost::format( "( %u ): Es wurden bei der Funktion zu wenige Paramteter angegeben." ) % error );
        break;

    case 4:
        text = boost::str( boost::format( "( %u ): Es wurden bei der Funktion zu viele Paramteter angegeben." ) % error );
        break;

    case 5:
        text = boost::str( boost::format( "( %u ): Geforderte Datenmenge wurde unterschritten." ) % error );
        break;

    case 6:
        text = boost::str( boost::format( "( %u ): Geforderte Datenmenge wurde überschritten." ) % error );
        break;

    case 7:
        text = boost::str( boost::format( "( %u ): Gesamt CRC des Packetes ist falsch.." ) % error );
        break;

    case 8:
        text = boost::str( boost::format( "( %u ): Es werden zusätzliche Steuerinformationen benötigt." ) % error );
        break;

    case 9:
        text = boost::str( boost::format( "( %u ): Die gesendete Protokollversion ist unbekannt." ) % error );
        break;

    case 10:
        text = boost::str( boost::format( "( %u ): Es wurde eine Anfrage erwartet, Antwort erhalten." ) % error );
        break;

    case 11:
        text = boost::str( boost::format( "( %u ): Es wurde eine Antwort erwartet, Anfrage erhalten." ) % error );
        break;

    case 12:
        text = boost::str( boost::format( "( %u ): Packetgroesse und damit die reservierte Puffergroesse ueberschreitet." ) % error );
        break;

    case 13:
        text = boost::str( boost::format( "( %u ): Entschluesselung des Packetinhaltes bringt Fehler." ) % error );
        break;

    case 14:
        text = boost::str( boost::format( "( %u ): Geraet arbeitet verschluesselt und Anfrage war unverschluesselt." ) % error );
        break;

    case 15:
        text = boost::str( boost::format( "( %u ): Bei der Ausführung ist ein interner Timeout eingetreten." ) % error );
        break;

    case 16:
        text = boost::str( boost::format( "( %u ): Angesprochener Slot ist nicht vorhanden.") % error );
        break;

    case 17:
        text = boost::str( boost::format( "( %u ): Kein Timeboy im Slot vorhanden.") % error );
        break;

    case 18:
        text = boost::str( boost::format( "( %u ): Timeboytyp noch nicht erkannt.") % error );
        break;

    case 19:
        text = boost::str( boost::format( "( %u ): Es wird versucht ein TimeboyI-III anzusprechen.") % error );
        break;

    case 20: // Dockingstation liest Datensätze vom Timeboy oder überträgt Listen.
        text = boost::str( boost::format( "( %u ): Ausführung konnte wegen belegtem Kommunikationskanal nicht weitergereicht werden.") % error );
        break;


    case 28:
        text = boost::str( boost::format( "( %u ): Das angegebene Passwort stimmt nicht mit dem benötigten überein. Es wird ein Passwort gefordert.") % error );
        break;

    case 29:
        text = boost::str( boost::format( "( %u ): Gerät ist aufgrund von fehlerhaften Passworteingaben für 60 Sekunden gesperrt.") % error );
        break;
    }

    if ( text.empty() )
    {
        text = boost::str( boost::format( "( %u ): Nicht abgefangener Modulfehler!" ) % error );
    }

    // Letzte Abfrage halten.
    m_error_module_number = error;
    m_error_module_text = text;
}


// ##################################################################
/**
* @brief Beschreibung fehlt noch.
*/
void Datafox::Communication::Module_base::error_function( uint16_t module, uint16_t func, uint16_t error, std::string &text )
{
    trace( "%s(%i):INFO(3): %s( %u, %u, %u, %p )\n", __FILE__, __LINE__, __func__, module, func, error, text.c_str() );

    // Sonderfehler für Passwortschutz abhandeln.
    switch ( error )
    {
        case 28:
        {
            // Die Modulfunktionen liefern den Fehlercode 28, wenn kein oder ein falsches Passwort verwendet wurde.
            // Es ist in diesem Fall zuerst das richtige Passwort für die Kommunikation erforderlich.
            text = boost::str( boost::format( "( %u ): Korrektes Passwort für die Kommunikation erforderlich.") % error );
            break;
        }

        case 29:
        {
            // Die Modulfunktionen liefern den Fehlercode 29, wenn 3 mal ein falsches Passwort verwendet wurde.
            // Die Befehlsausführung ist danach für 60 Sekunden blockiert.
            text = boost::str( boost::format( "( %u ): Gerät ist aufgrund von fehlerhaften Passworteingaben für 60 Sekunden gesperrt.") % error );
            break;
        }
    }

    if ( text.empty() )
    {
        text = boost::str( boost::format( "( %u ): Nicht abgefangener Funktionsfehler!" ) % error );
    }

    // Letzte Abfrage halten.
    m_error_function_number = error;
    m_error_function_text = text;
}

// ##################################################################
/**
* @brief Logen der textuellen Fehlerentsprechung.
*
*/
void Datafox::Communication::Module_base::trace_error( uint8_t device_address, uint16_t module, uint16_t func, int error )
{
    trace( "%s(%i):INFO(3): %s( %u, %u, %u )\n", __FILE__, __LINE__, __func__, device_address, module, func );

    std::string text;

    if ( error < 0 )
    {
        // Fehlertext ermitteln und eintragen.
        get_protocol().error_communication( -error, text );

        trace( "%s(%i):INFO(2): %s() [ %04X: %s ]\n", __FILE__, __LINE__, __func__, -error & 0xFFFF, text.c_str() );
    }
    else if ( error > 0 )
    {
        // Fehlertext ermitteln und eintragen.
        if ( IV_RESULT_MODULE(error) )
        {
            // Modulefehlertext ermitteln
            error_module( module, IV_RESULT_MODULE(error), text );

            trace( "%s(%i):INFO(2): %s() [ %04X: %s ]\n", __FILE__, __LINE__, __func__, error & 0xFFFF, text.c_str() );
        }
        else
        {
            // Funktionsfehlertext ermitteln
            error_function( module, func, IV_RESULT_FUNCTION(error), text );

            trace( "%s(%i):INFO(2): %s() [ %04X: %s ]\n", __FILE__, __LINE__, __func__, error & 0xFFFF, text.c_str() );
        }
    }
}

//#########################################################
//###################### HARDWARE #########################
//#########################################################

// ##################################################################
/**
* @brief Ausfuehrung einer Modulfunktion in ihrem jeweiligen Kontext.
*
*/
int Datafox::Communication::Module_base::call( uint16_t context, uint16_t module, uint16_t func, uint8_t *buffer, uint16_t &count, int context_value )
{
    trace( "%s(%i):INFO(3): %s( %u, %u, %u, %p, %u, %d )\n", __FILE__, __LINE__, __func__, context, module, func, buffer, count, context_value );

    return get_protocol().call( context, module, func, buffer, count, context_value );
}

// ##################################################################
/**
* @brief Beschreibung fehlt noch.
*
* @param deviceAddress Anzusprechende Geraetenummer.
*/
int Datafox::Communication::Module_base::remote_call( uint16_t module, uint16_t func, uint8_t *data, uint16_t &count )
{
    if ( module == IV_MODULE_BASE )
    {
        switch ( func )
        {
        case IV_FUNCTION_PING:

        // Funktion nicht definiert
        default:
            return 0x0200;
        }
    }

    // Fehler: Das geforderte Modul wird nicht unterstuetzt.
    return 0x0100;
}

// ##################################################################
/**
* @brief Prueft die Erreichbarkeit des Kommunikationspartners.
*
* @param deviceAddress Anzusprechender Kommunikationspartner.
*/
int Datafox::Communication::Module_base::ping( uint8_t device_address, uint8_t *data, uint16_t *count )
{
    trace( "%s(%i):INFO(3): %s( %u, %p, %u )\n", __FILE__, __LINE__, __func__, device_address, data, count );
    int result = IV_ERROR_SUCCESS;
    uint16_t lCount = 0;
    uint32_t auto_baud_rate[5] = { 38400, 9600, 115200, 57600, 19200 };
    uint8_t loop;

    // Funktion in Remoteklasse aufrufen?
    if ( !is_own_device_address( device_address ) )
    {
        // Befehlsausfuehrung geschuetzt durch Mutex.
        lock_guard_call guard( m_mutex_call );

        // Umgebung zuruecksetzen
        reset_error();

        // Remoteadresse setzen.
        get_protocol().set_device_address_receiver( device_address );

        // Zu sendende Daten kopieren?
        if ( data && count && *count > 0 )
        {
            lCount = *count;
            memcpy( m_packet_data, data, lCount );
        }

        Connection_base::Connection_type connection_type;
        std::string connection_string;
        uint32_t connection_value;
        uint32_t previous_connection_value;
        uint32_t communication_read_timeout;
        uint32_t previous_communication_read_timeout;
        get_connection().get_open_parameters( connection_type, connection_string, connection_value, communication_read_timeout );
        previous_connection_value = connection_value;
        previous_communication_read_timeout = communication_read_timeout;

        for ( loop = 0; loop < (sizeof( auto_baud_rate ) / sizeof( auto_baud_rate[0] )); ++loop )
        {
            // Bei Autobaud und Seriell wird die Kommunikation in folgender Reihenfolge abgefragt.
            // 1. Voreingestellter Baudwert.
            // Je nach Voreinstellung folgende Prio.
            // - 38400 bd
            // - 9600 bd
            // - 115200 bd
            // - 57600 bd
            // - 19200 bd
            if ( get_auto_baud() && connection_type == Connection_base::Connection_type::serial )
            {
                // Ab zweitem loop auf naechste Baudrate aus dem Pool wechseln.
                if ( loop > 0 )
                {
                    if ( auto_baud_rate[0] )
                    {
                        connection_value = auto_baud_rate[0];
                    }
                    else if ( auto_baud_rate[1] )
                    {
                        connection_value = auto_baud_rate[1];
                    }
                    else if ( auto_baud_rate[2] )
                    {
                        connection_value = auto_baud_rate[2];
                    }
                    else if ( auto_baud_rate[3] )
                    {
                        connection_value = auto_baud_rate[3];
                    }
                    else if ( auto_baud_rate[4] )
                    {
                        connection_value = auto_baud_rate[4];
                    }
                    else
                    {
                        break; // Abbruch der Schleife, wenn alle zu pruefenden durch sind.
                    }

                    // Schnittstellenbaud anpassen
                    get_connection().change_state_serial( get_connection().get_serial_port(), connection_value, communication_read_timeout );
                }

                // Aktuell gewaehlte Baudrate aus dem Pool nehmen.
                if ( connection_value == auto_baud_rate[0] )
                {
                    auto_baud_rate[0] = 0;
                }
                else if ( connection_value == auto_baud_rate[1] )
                {
                    auto_baud_rate[1] = 0;
                }
                else if ( connection_value == auto_baud_rate[2] ) 
                {
                    auto_baud_rate[2] = 0;
                }
                else if ( connection_value == auto_baud_rate[3] )
                {
                    auto_baud_rate[3] = 0;
                }
                else if ( connection_value == auto_baud_rate[4] )
                {
                    auto_baud_rate[4] = 0;
                }
            }
            else
            {
                // Kein Autobaud gewuenscht oder kein RS232.
                loop = (sizeof( auto_baud_rate ) / sizeof( auto_baud_rate[0] ));
            }

            if ( (result = call( IV_CONTEXT_MASTER, IV_MODULE_PROTOCOL, IV_FUNCTION_PING, m_packet_data, lCount, 2 )) == IV_ERROR_SUCCESS 
                || (result == -IV_ERROR_NEED_USER_PASSWORD) )
            {
                break;
            }
        }

        // Resultwert hier abpruefen!
        if ( result != IV_ERROR_SUCCESS )
        {
            // Nach Fehlgeschlagener Ermittlung, urspruengliche Einstellung wiederherstellen.
            if ( get_auto_baud() && connection_type == Connection_base::Connection_type::serial )
            {
                get_connection().change_state_serial( get_connection().get_serial_port(), previous_connection_value, previous_communication_read_timeout );
            }

            // Fehler ausgeben
            trace_error( device_address, IV_MODULE_PROTOCOL, IV_FUNCTION_PING, result );
            return result;
        }

        // Empfangene Daten kopieren?
        if ( count )
        {
            *count = lCount;
            if ( data && *count > 0 )
            {
                lCount = *count;
                memcpy( m_packet_data, data, lCount );
            }
        }
    }
    else
    {
        // Parameterdaten werden beibehalten

        // Antwort des Befehls senden.
        if ( (result = call( IV_CONTEXT_TRANSMIT_ANSWER, IV_MODULE_PROTOCOL, IV_FUNCTION_PING, data, lCount )) != IV_ERROR_SUCCESS )
        {
            trace( "%s(%i):ERROR(1): %s() [ call (%i) ]\n", __FILE__, __LINE__, __func__, result );
            return result;
        }
    }

    return result;
}

#ifdef WITH_VIRTUAL_DEVICE
//#########################################################
//############## ID und RESOURCEN #########################
//#########################################################

void Datafox::Communication::Module_base::set_uuid( boost::uuids::uuid &uuid )
{
    m_uuid = uuid;
}

boost::uuids::uuid Datafox::Communication::Module_base::get_uuid()
{
    return m_uuid;
}

std::string Datafox::Communication::Module_base::get_uuid_string()
{
    return boost::uuids::to_string( get_uuid() );
}

void Datafox::Communication::Module_base::set_path_cache( const boost::filesystem::path &path )
{
    m_path_cache = path;
}

boost::filesystem::path Datafox::Communication::Module_base::get_path_cache()
{
    return m_path_cache;
}

std::string Datafox::Communication::Module_base::get_path_cache_device()
{
    boost::filesystem::path path( get_path_cache() );
    path /= get_uuid_string();
    return path.string();
}

std::string Datafox::Communication::Module_base::get_file_path_device()
{
    boost::filesystem::path path( get_path_cache_device() );
    path /= "device.json";
    return path.string();
}

void Datafox::Communication::Module_base::load_from_cache()
{
    // Aktuellen Propertytree lesen.
    try
    {
        if ( boost::filesystem::exists( get_file_path_device() ) )
        {
            boost::property_tree::read_json( get_file_path_device(), m_device_properties );
        }
    }
    catch ( const boost::filesystem::filesystem_error& e )
    {
        trace( "filesystem_error( %s )\n", e.what() );
    }
    catch ( const boost::property_tree::json_parser_error& e )
    {
        trace( "json_parser_error( %s )\n", e.what() );
    }
}

void Datafox::Communication::Module_base::save_to_cache()
{
    // Aktuellen Propertytree schreiben.
    try
    {
        if ( !boost::filesystem::exists( get_path_cache_device() ) )
        {
            boost::filesystem::create_directories( get_path_cache_device() );
        }

        boost::property_tree::write_json( get_file_path_device(), m_device_properties );
    }
    catch ( const boost::filesystem::filesystem_error& e )
    {
        trace( "filesystem_error( %s )\n", e.what() );
    }
    catch ( const boost::property_tree::json_parser_error& e )
    {
        trace( "json_parser_error( %s )\n", e.what() );
    }
}
#endif
